import datetime
import logging
import uuid

import celery.result
from celery import states

logger = logging.getLogger(__name__)


def _serialize(value):
    if isinstance(value, datetime.datetime):
        return value.isoformat()
    if isinstance(value, dict):
        return {key: _serialize(key_value) for key, key_value in value.items()}
    if isinstance(value, list):
        return [_serialize(list_value) for list_value in value]
    if isinstance(value, tuple):
        return (_serialize(tuple_value) for tuple_value in value)
    if isinstance(value, set):
        return {_serialize(set_value) for set_value in value}
    return value


def _async_result_stub(task_id, **kwargs):
    return _TaskResultStore.get_by_id(task_id)


# apply_async returns an EagerResult in eager mode.
# To ensure it always returns an EagerResult even when AsyncResult is called, we use this mock
celery.result.AsyncResult = _async_result_stub


class _EagerResultWithStateSupport(celery.result.EagerResult):
    def ready(self):
        return self._state == states.READY_STATES


class _TaskResultStore:

    __task_store = {}

    @classmethod
    def put(cls, result):
        cls.__task_store[result.id] = result

    @classmethod
    def get_by_id(cls, id):
        return (
            cls.__task_store[id]
            if id in cls.__task_store
            else _EagerResultWithStateSupport(id, None, states.PENDING)
        )


class CeleryMock:
    """
    Celery App proxy. This proxy configures celery app in "task always eager" mode.
    This proxy intercepts task decorator so apply_async is working in this mode.
    """

    def __init__(self, celery_app):
        self.__celery_app = celery_app
        self.__celery_app.conf.update(
            task_always_eager=True,
            result_backend="cache",
            cache_backend="memory",
            task_eager_propagates=True,
        )
        self.__task_store = {}

    def __getattr__(self, name):
        if name == "task":

            def task_interceptor(*aa, **oo):
                result = getattr(self.__celery_app, "task")(*aa, **oo)

                class AsyncTaskProxy:
                    def __init__(self, method, *a, **o):
                        self.__method = method

                    def apply_async(
                        self,
                        args=None,
                        kwargs=None,
                        task_id=None,
                        producer=None,
                        link=None,
                        link_error=None,
                        shadow=None,
                        **options,
                    ):
                        serialized_args = (
                            [_serialize(arg) for arg in args] if args else ()
                        )
                        serialized_kwargs = (
                            {name: _serialize(value) for name, value in kwargs.items()}
                            if kwargs
                            else {}
                        )

                        task_id = task_id if task_id else str(uuid.uuid4())

                        try:
                            method_result = self.__method(
                                *serialized_args, **serialized_kwargs
                            )
                            celery_result = celery.result.EagerResult(
                                task_id, method_result, states.SUCCESS
                            )
                        except Exception as e:
                            celery_result = celery.result.EagerResult(
                                task_id, e, states.FAILURE
                            )

                        _TaskResultStore.put(celery_result)
                        return celery_result

                    def __call__(self, *args, **kwargs):
                        return self.__method(*args, **kwargs)

                return AsyncTaskProxy

            return task_interceptor
        else:
            return getattr(self.__celery_app, name)



import logging
import os

from huey import RedisHuey
from huey.exceptions import TaskException


logger = logging.getLogger("asynchronous_server")

# List all statements that should be executed before evaluating an Exception that occurred.
imported_exceptions = []


def build_async_application(config: dict, **kwargs) -> RedisHuey:
    """
    This function should be called within a python module called 'asynchronous_server.py'
    Build a huey application with given configuration and modules
    :param config: Dictionary with following structure: {
        'asynchronous': {
            'broker': ...,
        }
    }
    :param kwargs: Additional Huey arguments
    :return: RedisHuey Application
    """
    logger.info(f"Starting Huey server")
    return RedisHuey(
        os.getenv("CONTAINER_NAME", "LOCAL"),
        url=config["asynchronous"]["broker"],
        **kwargs,
    )


def _get_asynchronous_task(huey_task_id: str, huey_app: RedisHuey):
    try:
        return huey_app.result(huey_task_id, preserve=True)
    except:
        return "Exception"


def _result_is_available(huey_task):
    return huey_task is not None


def _get_current_state(huey_task):
    # TODO Add more information such as request initial time, and maybe intermediate client status
    return "PENDING"


def _get_asynchronous_result(huey_app: RedisHuey, huey_task_id: str):
    try:
        huey_task = huey_app.result(huey_task_id)
    except TaskException as e:
        try:
            for import_exception in imported_exceptions:
                exec(import_exception)
            task_exception = eval(e.metadata["error"])
        except NameError:
            task_exception = Exception(e.metadata["error"])
        raise task_exception
    return huey_task



from datetime import datetime
import logging
import os

import celery.result
from celery import Celery, current_task
from celery.task import control

logger = logging.getLogger("asynchronous_server")


def build_async_application(config: dict, *apps: str, **kwargs) -> Celery:
    """
    This function should be called within a python module called 'celery_server.py'
    Build a celery application with given configuration and modules
    :param config: Dictionary with following structure: {
        'celery': {
            'broker': ...,
            'backend': ...,
        }
    }
    :param apps: celery application modules
    :param kwargs: Additional Celery arguments
    To add celery configuration parameter, you should provide a dictionary named changes with those parameters.
    As in changes={'task_serializer': 'pickle'}
    :return: Celery Application
    """
    namespace = _namespace()
    queue = _queue()

    logger.info(f"Starting Celery server on {namespace} namespace")

    return Celery(
        "celery_server",
        broker=config["celery"]["broker"],
        # Store the state and return values of tasks
        backend=config["celery"]["backend"],
        namespace=namespace,
        include=apps,
        changes={
            "worker_hijack_root_logger": False,
            # Use pickle instead of msgpack for now as there is a bug in Celery 4.3.0 with redis
            "task_serializer": "pickle",
            "result_serializer": "pickle",
            "accept_content": ["pickle"],
            "task_compression": "gzip",
            "task_default_queue": queue,
            "task_default_exchange": queue,
            "task_default_routing_key": queue,
            **kwargs.pop("changes", {}),
        },
        **kwargs,
    )


def _get_asynchronous_task(celery_task_id: str, celery_app: Celery):
    return celery.result.AsyncResult(celery_task_id, app=celery_app)


def _result_is_available(celery_task):
    return celery_task.ready()


def _get_current_state(celery_task):
    # TODO Add more information such as request initial time, and maybe intermediate client status
    return celery_task.state


def _get_asynchronous_result(celery_app: Celery, celery_task_id: str):
    celery_task = celery.result.AsyncResult(celery_task_id, app=celery_app)
    return celery_task.get(propagate=True)


def _namespace() -> str:
    """
    Workers are started using CONTAINER_NAME environment variable as namespace or local.
    Followed by a unique identifier per machine (HOSTNAME environment variable or localhost)
    """
    return (
        f"{os.getenv('CONTAINER_NAME', 'local')}_{os.getenv('HOSTNAME', 'localhost')}"
    )


def _queue():
    """
    Workers are started using CONTAINER_NAME environment variable as queue or local.
    """
    return os.getenv("CONTAINER_NAME", "local")


def health_details():
    try:
        worker_name = f"celery@{_namespace()}"
        workers = control.ping(destination=[worker_name])
        if not workers:
            return (
                "fail",
                {
                    "celery:ping": {
                        "componentType": "component",
                        "status": "fail",
                        "time": datetime.utcnow().isoformat(),
                        "output": f"No {worker_name} workers could be found.",
                    }
                },
            )

        return (
            "pass",
            {
                "celery:ping": {
                    "componentType": "component",
                    "observedValue": workers,
                    "status": "pass",
                    "time": datetime.utcnow().isoformat(),
                }
            },
        )
    except Exception as e:
        return (
            "fail",
            {
                "celery:ping": {
                    "componentType": "component",
                    "status": "fail",
                    "time": datetime.utcnow().isoformat(),
                    "output": str(e),
                }
            },
        )


class CeleryTaskIdFilter(logging.Filter):
    """
    This is a logging filter that makes the celery task identifier available for use in the logging format.
    This filter support lookup in celery context for the current task id
    """

    def filter(self, record):
        record.celery_task_id = (
            current_task.request.id
            if current_task
            # TODO Ensure that those checks really make sense
            and hasattr(current_task, "request") and hasattr(current_task.request, "id")
            else ""
        )
        return True



import os
from setuptools import setup, find_packages

this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.md"), "r") as f:
    long_description = f.read()

setup(
    name="flasynk",
    version=open("flasynk/version.py").readlines()[-1].split()[-1].strip("\"'"),
    author="Colin Bounouar",
    author_email="colin.bounouar.dev@gmail.com",
    maintainer="Colin Bounouar",
    maintainer_email="colin.bounouar.dev@gmail.com",
    url="https://colin-b.github.io/flasynk/",
    description="Exposing Asynchronous endpoint using Flask-RestPlus",
    long_description=long_description,
    long_description_content_type="text/markdown",
    download_url="https://pypi.org/project/flasynk/",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Build Tools",
    ],
    keywords=["async", "asynchronous", "rest", "flask"],
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        # Used to manage endpoints and open api definition
        "flask-restplus==0.13.*"
    ],
    extras_require={
        "celery": ["celery[redis,msgpack]==4.*"],
        "huey": ["huey==2.*", "redis==3.*"],
        "testing": [
            # Extra requirements
            "celery[redis,msgpack]==4.*",
            "huey==2.*",
            # Used to manage testing of a Flask application
            "pytest-flask==0.15.*",
            # Used to check coverage
            "pytest-cov==2.*",
        ],
    },
    python_requires=">=3.7",
    project_urls={
        "GitHub": "https://github.com/Colin-b/flasynk",
        "Changelog": "https://github.com/Colin-b/flasynk/blob/master/CHANGELOG.md",
        "Issues": "https://github.com/Colin-b/flasynk/issues",
    },
    platforms=["Windows", "Linux"],
)



import json
import logging
import re
from urllib.parse import urlparse

import flask
from flask_restplus import Resource, fields, Namespace


logger = logging.getLogger("asynchronous_server")

_STATUS_ENDPOINT = "status"
_RESULT_ENDPOINT = "result"


class AsyncNamespaceProxy:
    """
    Flask rest-plus Namespace proxy.
    This proxy namespace add a decorator 'async_route' that will generate 2 extra endpoint : /status and /result
    to query the status or the result of the huey task
    """

    def __init__(self, namespace: Namespace, async_app):
        self.__namespace = namespace
        self.__async_app = async_app
        task_status_model = namespace.model(
            "AsyncTaskStatusModel",
            {
                "task_id": fields.String(required=True, description="Task Id."),
                "url": fields.String(
                    required=True, description="URL when task status can be checked."
                ),
            },
        )
        self.how_to_get_asynchronous_status_doc = {
            "responses": {
                202: (
                    "Computation started.",
                    task_status_model,
                    {"headers": {"location": "URL to fetch computation status from."}},
                )
            }
        }

    def __getattr__(self, name):
        return getattr(self.__namespace, name)

    def asynchronous_route(self, endpoint: str, serializer=None, to_response=None):
        """
        Add an async route endpoint.
        :param endpoint: value of the exposes endpoint ex: /foo
        :param serializer: In case the response needs serialization, a single model or a list of model.
        If a list is given, the output will be treated (serialized) and documented as a list
        :param to_response: In case the task result needs to be processed before returning it to client.
        This is a function taking the task result as parameter (and path parameters if needed) and returning a result.
        Default to returning unmodified task result.
        :return: route decorator
        """

        def wrapper(cls):
            # Create the requested route
            self.__namespace.route(endpoint)(cls)
            # Create two additional endpoints to retrieve status and result
            _build_result_endpoints(
                cls.__name__,
                endpoint,
                self.__namespace,
                self.__async_app,
                serializer,
                to_response,
            )
            return cls

        return wrapper


def _base_url() -> str:
    """
    Return client original requested URL in order to make sure it works behind a reverse proxy as well.
    Without parameters.
    """
    if "X-Original-Request-Uri" in flask.request.headers:
        parsed = urlparse(flask.request.headers["X-Original-Request-Uri"])
        return f'{flask.request.scheme}://{flask.request.headers["Host"]}{parsed.path}'
    return flask.request.base_url


def how_to_get_asynchronous_status(async_task) -> flask.Response:
    url = f"{_base_url()}/{_STATUS_ENDPOINT}/{async_task.id}"
    status = flask.Response()
    status.status_code = 202
    status.content_type = "application/json"
    status.headers["location"] = url
    status.data = json.dumps({"task_id": async_task.id, "url": url})
    return status


def _get_asynchronous_status(async_task_id: str, async_app) -> flask.Response:
    async_module = _module(async_app)
    async_task = async_module._get_asynchronous_task(async_task_id, async_app)
    if async_module._result_is_available(async_task):
        status = flask.Response()
        status.status_code = 303
        status.headers["location"] = _base_url().replace(
            f"/{_STATUS_ENDPOINT}/", f"/{_RESULT_ENDPOINT}/"
        )
        return status

    # TODO Add more information such as request initial time, and maybe intermediate client status
    return flask.jsonify({"state": async_module._get_current_state(async_task)})


def _build_result_endpoints(
    base_class,
    endpoint_root: str,
    namespace: Namespace,
    async_app,
    response_model,
    to_response: callable,
):
    @namespace.route(f"{endpoint_root}/{_RESULT_ENDPOINT}/<string:task_id>")
    class AsyncTaskResult(Resource):
        @_conditional_marshalling(namespace, response_model)
        @namespace.doc(f"get_{_snake_case(base_class)}_result")
        def get(self, task_id: str, **kwargs):
            """
            Retrieve result for provided task.
            """
            result = _module(async_app)._get_asynchronous_result(async_app, task_id)
            return to_response(result, **kwargs) if to_response else result

    @namespace.route(f"{endpoint_root}/{_STATUS_ENDPOINT}/<string:task_id>")
    @namespace.doc(
        responses={
            200: (
                "Task is still computing.",
                namespace.model(
                    "CurrentAsyncState",
                    {
                        "state": fields.String(
                            description="Indicates current computation state.",
                            required=True,
                            example="PENDING",
                        )
                    },
                ),
            ),
            303: (
                "Result is available.",
                None,
                {"headers": {"location": "URL to fetch results from."}},
            ),
            500: (
                "An unexpected error occurred.",
                fields.String(description="Stack trace.", required=True),
            ),
        }
    )
    class AsyncTaskStatus(Resource):
        @namespace.doc(f"get_{_snake_case(base_class)}_status")
        def get(self, task_id: str, **kwargs):
            """
            Retrieve status for provided task.
            """
            return _get_asynchronous_status(task_id, async_app)


def _conditional_marshalling(namespace: Namespace, response_model):
    def wrapper(func):
        if response_model is not None:
            return namespace.marshal_with(
                response_model[0]
                if isinstance(response_model, list)
                else response_model,
                as_list=isinstance(response_model, list),
            )(func)
        return func

    return wrapper


def _snake_case(name: str) -> str:
    if "_" in name:
        raise ValueError(f"{name} should be Camel Case and should not contain any _")
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def _module(async_app):
    try:
        import huey

        if isinstance(async_app, huey.api.Huey):
            import flasynk.huey_specifics

            return flasynk.huey_specifics
    except ModuleNotFoundError:
        pass  # If huey cannot be imported then it means Celery is used
    import flasynk.celery_specifics

    return flasynk.celery_specifics
